#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 19 15:19:29 2020
file split on Sun Oct 18 16:07:53 2020
@author: davidworrall
version 0.01: Fri Jun 19 15:19:29 2020. Set up intitial cv read
version 0.02: 2020 31 Define class structure for a country
version 0.03: 2020 08 06 Add SMA
version 0.04: 2020 09 18 finish marked-line plotting routines * non-printing vals <- 0
version 0.05: 2020 09 19 finish x-axis 7day divisions and labels
version 0.06: 2020 09 23 regional sort according to classification in WorldRegionsUpdated.csv
            complete plot generation and write to file
version 0.07 begin investigation of auto anomoly detection. found TAD (twitter ...)
see https://blog.twitter.com/engineering/en_us/a/2015/introducing-practical-and-robust-anomaly-detection-in-a-time-series.html

version 0.08 code Split code into 3 files:
    - covid19data.py  .... included here
    - covid19class.py .... (this file)
    -covid19plotting.py .. includes all of the above

databases in play:
  - deathsDB = deathsDB[1:] # drop the column headings
  - dates = datePad + columnLabels[4:]
  - countriesList = []      # collates all multiple-province country provinces into a single country entry
  - provinceDeathsDict = {} # holds all country, province and summed-province data
  - populationDataList = []       # demographic data CONVER TO DICT

  - inJustDeathDB = []      # countries/provinces that are only in deaths_global.csv
  - inJustpopulationDataList = [] # countries/provinces that are only in countryPopulationData.csv
  - inBothDB = []
  - demographicDict = {}    # an amalgulm of the two .csv files data # regionName,countryName,countrySymbol,area,pop

""" 
from demographicData import *   # this also  imports covid19data.py

# list of provinces with outlier anomolies to adjust before plotting. Actual adjustment are donein plotting routines
outlierProvinceList = ['Argentina', 'Bolivia', 'China', 'Equador', 'India', 
                       'Ireland', 'Kyrgzstan', 'Mexico', 'Pakistan', 'Peru', 'Sudan'] 

# ====================DEFINE DEATH CLASS FOR GLOBAL COUNTRIES/STATES?PROVINCES#

class CountryDeaths():    # define a class for handling Covid-19 death data for a country
    def __init__(self, countryName):
        #Initialize data attributes.
        self.countryName        = countryName
        self.countryExists      = False
        self.printData          = False                                         # used to filter death data when using print (instance)
        self.provinceName       = ""                                            # filled from deahDB  by __init__()
        self.region             = ""                                            # filled from populationDataList  by __init__()
#        self.latitude           = countriesList[latCol_deathsData]
#        self.longitude          = countriesList[latCol_deathsData]
        if self.countryName in demographicDict.keys():                                       # defined in demographicData.py
            self.countryCode     = demographicDict [self.countryName][countrySymbolCol]      # Unique two letter code
            self.area            = demographicDict [self.countryName][areaCol]               # land area (in sq kms)
            self.population      = int(demographicDict [self.countryName][populationCol])    # the population
            if self.population > 0:
                self.populationDensity  = self.population/self.area
                                    #sun     mon       tuea          wed
        self.prePadDates        = ['1/19/20','1/20/20', '1/21/20']              # 1/22/20 = first data date. prepad for a Sunday start
        self.prePadData         = [0] * len (self.prePadDates)                  # insert len ( self.prePadDates days of data = 0)
        self.cumulativeDeaths   = self.prePadData                               # start the list by prepadding
        self.deathIncrements    = []                                            # NB calculated from self.cumulativeDeaths, so no need to prePad
        self.ascending          = []
        self.sortedIndices      = []                                            # indices of increments in ascending order
        self.NrDataPoints       = 0                                             # Needed many times so no need to keep calculating it
        self.mean               = 0                                             # Mean nr of Deaths / day calculated in the Stats method
        self.median             = 0                                             # middle nr of Deaths / day calculated in the Stats method
        self.max                = 0                                             # Max Nr of Deaths in any one day calculated in the Stats method
        self.SMAshortPeriod     = 4                                             # calculated in the Stats method
        self.SMAlongPeriod      = 14                                            # calculated in the Stats method
        self.SMAshortData      = []                                             # Simple Moving Average over self.SMAperioShort  
        self.SMAlongData        = []                                             # Simple Moving Average over self.SMAperiods  
        self.anomolies2adjust   = 0                                             # non-zero if adjusted
        self.maxVals            = ()                                            #indices of the 3 largest daily death values # fix the values

        if self.countryName not in provinceDeathsDict.keys():
            self.countryExists = False
            print (self.countryName + " does not exist in the Database.\nIt contains data for the following countries:")
 #           kees = sorted (provinceDeathsDict.keys())                          # alpha sort for 'pretty' print
 #           for keeNr in range (len (kees)):
 #                print (kees[keeNr], end="\t")
 #               if keeNr % 3 == 0: print()                                     # throw a CR
            return (None)
        else:
            self.countryExists = True
#            print("*********", provinceDeathsDict[self.countryName])
#            print
            self.provinceName   = provinceDeathsDict[countryName][0]            #countriesList[self.countryName][0]
            self.latitude       = provinceDeathsDict[countryName][2]            # countriesList[latCol_deathsData]
            self.longitude      = provinceDeathsDict[countryName][3]            # countriesList[longCol_deathsData]
            self.area           = None
            self.population     = None
            self.cumulativeDeaths = self.prePadData + provinceDeathsDict[countryName][4:]
            # block the printing of values <=0
            for indx in range (len(self.cumulativeDeaths)):
                if self.cumulativeDeaths[indx] <= 0:
                    self.cumulativeDeaths[indx] = 0 # float ('nan')             # nan = not-a-number
            self.deathIncrements = [0]* len (self.cumulativeDeaths)
            for dayNr in range (1,len (self.cumulativeDeaths)):
                self.deathIncrements[dayNr] = self.cumulativeDeaths[dayNr]-self.cumulativeDeaths[dayNr-1]
                if self.deathIncrements[dayNr] <0:                              # indicates data error
                    if self.printData:
                        print ("DATA ERROR:",self.countryName , end="\t")
#                      print ("DATA ERROR:Some dead rose again in ",self.countryName , "on", :", end="/t")
                        print (dayNr,":",dates[dayNr], "\t", self.cumulativeDeaths[dayNr-1],"\t",self.cumulativeDeaths[dayNr],"\tdiff :", self.deathIncrements[dayNr-1],self.deathIncrements[dayNr])

                    self.deathIncrements[dayNr] = 0  # just zero it : #  ============ ERROR CORRECTION==============
 #           self.generateStats()
            #  ============== NBNBNBNBNBNBN
            #print (self.deathIncrements)
            if self.provinceName == "":
                self.provinceName = self.countryName                            # the 'province' is the country
            self.NrDataPoints = len (self.deathIncrements)
            self.generateStats()                                                # execute before convertsion of 0's to NANS
            self.fixAnomalies()
#            self.zeros2NANs()                                                   # disable printing of 0's
   
#            countriesInConsiderationList += [self.countryName]

    def __str__(self):
#        # Defines what to do when the default python help() function is called on an instance
        if self.countryExists == True:
            print ('#' * 30)
            print ('Country:',                      self.countryName, end = ' ')
            print ('(Code: ',                       self.countryCode, ')')
#            print ('#' * 30)
            if self.provinceName !='' or (self.province == self.countryName) :
                print ('Province:\t' ,              self.provinceName)
            print (" Total Number of Deaths:\t",   self.cumulativeDeaths[-1])
            print ('µ (mean) of new deaths/day:\t', self.mean)
            print ('Max Nr of deaths/day:\t',       self.max )
            if self.countryName in demographicDict.keys():
                self.area               = demographicDict [self.countryName][areaCol]
                self.population         = demographicDict [self.countryName][populationCol]
                print ("CountryCode:\t",            self.countryCode)
                print ('Area (sq kms):\t',          self.area)
                print ('Population:\t' ,            self.population)
                if self.population > 0:
                    self.populationDensity  =       self.population/self.area
                    print ('Population Density:\t', self.populationDensity)
                else: print ("No demographic data available")
            if not self.printData:
                print ("Switch self.printData on for more complete details of "+self.countryName+ 'statistics.')
            if self.printData:
                print ('Latitude:\t' ,                  self.latitude)
                print ('Longitude:\t' ,                 self.longitude)
                print (self.countryName+': Cumulative Deaths:\n',       self.cumulativeDeaths)
                print (self.countryName+': Death Increments:\n',        self.deathIncrements)
                print (self.countryName+': Sorted deathIncrements:\n',  self.ascending)
        return ("")

    def generateStats (self):
        # generates some stats of the deaths. Perhaps run it twice - before and after fixing anomalies
        from statistics import mean, median
        self.median = int( median (self.deathIncrements))                       # not affected by leading zeros
        if any(self.deathIncrements):                                           # calc. only from the first non-zero day   
            indxof1stNonZero = next((i for i, x in enumerate(self.deathIncrements) if x), None) # find the index
            tmpIncrements = self.deathIncrements[indxof1stNonZero:]
            self.mean = int(round (mean(tmpIncrements)))            
#        self.mean = int(round (mean (self.deathIncrements)))
#        self.mean       = self.nanMean (self.deathIncrements) #; print ('Daily deaths, mean   :', self.mean)
        self.max        = max (self.deathIncrements) #; print ('Daily deaths, maxVal :', self.max )
        self.ascending  = sorted(self.deathIncrements)
#        sorted(range(len(vals)), key=vals.__getitem__)
        kee = self.deathIncrements.__getitem__
        deathIncrementIndices = range(len(self.deathIncrements))                # 0,1, 2, ... len()
        self.sortedIndices = sorted(deathIncrementIndices, key=kee)
#       alternative: sorted(range(len(vals)), key=vals.__getitem__)

    def valueIndexSort(self, listy):
        # ascending sort of listy. returns key value pairs
        sortedListy = sorted (listy)                                            # sort the items
        kee = listy.__getitem__                                                 # get the kee position 
        indices = range(len(self.deathIncrements))                              # 0,1, 2, ... len() 
        self.sortedIndices = sorted(deathIncrementIndices, key=kee)             # the sorted indices
        
    def fixAnomalies(self):
         if self.countryName in outlierProvinceList: 
             print ("Anomolie to fix: ", self.countryName, end = "\t")
             print ("[indexes, values]:", self.sortedIndices[-3:], self.ascending [-3:] )
             self.maxVals = self.sortedIndices[-3:]
             for indx in range(len(self.maxVals)): 
                 self.maxVals [indx] = (self.maxVals [indx],self.deathIncrements[self.maxVals [indx]] )
#            print (self.maxVals)   #indices of the 3 largest daily death values # fix the values
    
    def SMA (self, NrPeriods):
        #Computes a Simple Moving Average of self.deathIncrements over NrPeriods
        self.SMAperiod = NrPeriods
        self.SMAdata = []
        firstDataPointIndex = len (self.deathIncrements)                       # find first non-zero  datapoint
        lastDataPointIndex = len (self.deathIncrements)
        if lastDataPointIndex <  self.SMAperiod:
            print ("Averaging period too long." +
                   "Longest period is: ", lastDataPointIndex-self.SMAperiod)
            return(None)
        index = 0
        while index < self.NrDataPoints:
            thisWindow = self.deathIncrements[index: index+ self.SMAperiod]
            windowAverage = sum(thisWindow) / self.SMAperiod
#            print (thisWindow, windowAverage )
            self.SMAdata += [windowAverage]                                     # .append (windowAverage)
            index += 1

    def SMA2(self, dataList,nrPeriods):
        # returns the simple moving average of dataList over nrPeriods
        import numpy as np
        cumSum = np.cumsum(np.insert(dataList, 0, 0))                           # cumulative sum of dataList
        return (cumSum[nrPeriods:] - cumSum[:-nrPeriods]) / float(nrPeriods)   # SMA over nrPeriods

    def nanMean (self, listy):
        #Computes the mean of a list containing float ('nan') values
        nrItems = len (listy)
        sumItems = 0
        for item in listy:
            try:
                sumItems += item                                                # cummulative sum of summable values in listy
            except:
                pass                                                            # faster than adding 0 used to avoid float ('nan' values for e.g.
#        print ("sum + len + mean", sumItems, len (listy), sumItems/len(listy))
        return sumItems/len (listy)
    # median to b calculated also.           

    def zeros2NANs (self):
        # replace all zeros with float ('nan') [Not A Number] to suppress printing of all 0 values
        nrDays = len(self.deathIncrements)
        for indx in range (1, nrDays):                                          # don't use the first one - why?
            if self.deathIncrements[indx]>(200+self.deathIncrements[indx-1]):   # attempt at filtering outliers
                pass                                                            #  not working; not used
            # print (self.deathIncrements[indx-1], self.deathIncrements[indx])  
            #self.deathIncrements[indx] = 0 # self.deathIncrements[indx-1]
        for indx in range (nrDays):                                             # leave last value to set manually
            if self.deathIncrements[indx] <= 0:                                 # zero -ve increments:resurrection currently not available :-)
               self.deathIncrements[indx] = float('nan')
        self.deathIncrements[0]=0                                               # Kludge: force a printable val so graph will start at the 0th index
        if not (self.deathIncrements [-1] > 0):
            self.deathIncrements [-1] = float('nan')                            # Kludge: force a printable val so graph will end  at the 0th index
#            self.deathIncrements += [0]                                        # Kludge: force a printable val so graph will end  at the 0th index

# ===================  </End class CountryDeaths() definition> ====================

# ===================  INSTANTIATION OF PROVINCES (GLOBAL) ========================
# Instantiate all non-zero cases Provinces, make a list of the zero case Provinces

""" 2020 11 12 added:
 Instantiate here using listOfCountryNames, covidInstanceDict
"""
deathFreeProvinces = []

covidInstanceDict2 = {}                                                         # used to  store country instances for plotting etc

for province in provinceDeathsDict.keys():
    instance = CountryDeaths(province)
    if instance.cumulativeDeaths[-1] > 0:                                       # there are  deaths
        covidInstanceDict2[province] = instance
        instance.SMAshortPeriod = 4
        instance.SMAlongPeriod = 14
        print (instance.deathIncrements)
#        x = instance.SMA2(instance.deathIncrements,instance.SMAshortPeriod)
#        print (x)
        instance.SMAshortData = instance.SMA2(instance.deathIncrements,instance.SMAshortPeriod)
#        y = instance.SMA2(instance.deathIncrements,instance.SMAshortPeriod)
#        print (y)
        instance.SMAlongData = instance.SMA2(instance.deathIncrements,instance.SMAshortPeriod)
       
#        print (instance.SMAshortData)
#        print (instance.SMAlongData)

    else:                                                                       # there are NO deaths in this province
        deathFreeProvinces += [province]

    
    

#    provinceDeathsDict[province][0]     
#    instanceString = "covidInstanceDict['"+province+"'] = CountryDeaths('"+province+"')"
#    exec(instanceString)                  # create the instance
"""
# ============ PRINT PROVINCE DATA WITH MAX VALS & MAX VAL INDICES ===========
# print ("All Provinces\n=============\n", covidInstanceDict2.keys())
for kee in covidInstanceDict2.keys():
#    print (kee, end='\t')
    if kee in outlierProvinceList:
        maxVals = covidInstanceDict2[kee].sortedIndices[-3:]
        for indx in range(len(maxVals)):
            maxVals [indx] = (maxVals [indx],covidInstanceDict2[kee].deathIncrements[maxVals [indx]] )
print ()
#        print (maxVals)   #indices of the 3 largest daily death values # fix the values
#        print (covidInstanceDict2[kee].deathIncrements[])
#    else:
#        print ()

print ("Death-Free Provinces\n====================\n", deathFreeProvinces)

"""